/*------------
		    210 GHz Site Survey Radiometer
	       Five College Radio Astronomy Observatory
		University of Massachusetts at Amherst

			  David Hiriart & Luis Salas
				Sept-1992
				Nov -1992
                               (DOS Version)
			  David Hiriart & Douglas McGonagle
				Oct-94
                               (SUN Version)
			  David Hiriart 
                                Sept-2001
                               (Lynux Version)

  Oct 2002: Change tipper device from ttyS0 to tipper (TIPPER_DEV), and other 
            cosmetics changes.
	    Add trap for SIGTERM (Termination signal 15) over Break Handler
	    run-time.
	    Add syslog facilities.
	    Add verbose and debug modes on compiled time and some macros.
	    (Federico Bareilles)

						------------------*/

#define VERSION "0.2"

#ifdef DEBUG_MODE
# ifndef VERBOSE_MODE
#   define VERBOSE_MODE
# endif 
#endif

#include <ctype.h> /* CBAUD is not POSIX, it is SVID and BSD */
#define _POSIX_SOURCE 1         /* POSIX compliant source  */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <termio.h>
#include <signal.h>
#include <unistd.h> 
#include <getopt.h>
#include <syslog.h>
#include <errno.h>

#define IGREG   (15+31L*(10+12L*1582))
#define TIPPER_DEV "/dev/tipper"
#define DATA_DIR "/data"

#undef PVERB
#ifdef VERBOSE_MODE
#  define PVERB(fmt, args...) if(verbose_flag) fprintf(stderr, fmt, ## args)
#else 
#  define PVERB(fmt, args...)
#endif

#undef SYSLOG
#define SYSLOG(type, fmt, args...) if(log_flag) syslog(type, fmt, ## args)

#undef PERROR
#define PERROR(s) { PVERB("[ERROR: %s: %s]\n", (s),strerror(errno));\
                    SYSLOG(LOG_ERR,"%s: %s",(s), strerror(errno));\
                  }

#define EXIT(n) closelog(); exit(n)

#undef READ
#ifdef DEBUG_MODE
#  define READ(fd, buf, count) read(fd, buf, count);\
          if(character_flag){\
            int i;\
            for( i=0;i<count;i++) fprintf(stderr,"%c",*buf);\
          }
#else
#  define READ(fd, buf, count) read(fd, buf, count)
#endif


void cbrk_handler( int s);
long julday(int mm, int id,int iyyy);

struct termio save, term;
FILE *fp, *fps, *fpo;
char *program_name;

/* Flags: */
int verbose_flag      = 0;
int daemon_flag       = 1;
int log_flag          = 0;
int character_flag    = 0;

char *help_text[] ={
#ifdef VERBOSE_MODE
  "  -v\t\t\t- verbose mode.",
#endif
  "  -n, --nodaemon\t- no run as a daemon.",
  "  -l,[--log]\t\t- write events to syslog.",
#ifdef DEBUG_MODE
  "  -c,[--character]\t- print in stderr all characters arrived from serial\n\t\t\t  line.",
#endif
  "  -d\t\t\t- set data directoy (def.: "DATA_DIR")",
 "  --help\t\t- this help !!!.",
  ""
};

void usage()
{
  /*cartel();*/
  fprintf(stderr,
#ifdef VERBOSE_MODE
          "Usage: \t%s [-vnlh][--nodaemon][--log][--help]\n",
#else
          "Usage: \t%s [-nlh][--nodaemon][--log][--help]\n",
#endif
          program_name);
}

void help()
{
  int i=0;

  while((char) *help_text[i] != 0)
    fprintf(stderr,"%s\n",help_text[i++]);
 fprintf(stderr,"\n");
}


int main(int argc, char **argv)
{
  int  day,daysent,firtim;
  char c, *name=NULL, *data_dir=DATA_DIR;
  long int refday;
  int  mm,dd,yy,hh,min,sec;
  
  time_t now;
  struct tm *tmstruct;
  int pid;
  char *cp;
  
  int op;
  static struct option  const  longopts[] =
  {
    {"help", 0, 0, 101},
    {"nodaemon", 0, 0, 'n'},
    {"log", 0, 0, 'l'},
    {0,  0,  0,  0}
  };
  
  /* who is my name? */
  if ((cp = strrchr(argv[0], '/')) != NULL) program_name  = cp + 1;
  else program_name  = argv[0];
  
  opterr =  0;
  while  ((op = getopt_long(argc, argv, "vnlcd:",longopts, 0)) != EOF)
    switch( op ){
    case 'v':
      verbose_flag = 1;
      break;
#ifdef DEBUG_MODE
    case 'c':
      character_flag = 1;
      break;
#endif
    case 101:
      usage();
      help();
      exit(-1);
    case 'n':
      daemon_flag = 0;
      break;
    case 'l':
      log_flag = 1;
      break;
    case 'd':
      data_dir=optarg;
      break;
    default:
      usage();
      exit(-1);
    }
  if(daemon_flag) verbose_flag = 0;

  if(daemon_flag){
    /* Ok, a daemond */
    if ( (pid=fork()) != 0 ) {
      exit(0);
    }
  }

  if(log_flag){
    openlog(program_name,LOG_PID, LOG_DAEMON );
    SYSLOG(LOG_NOTICE,"Startup succeeded.");
  }

  /*  Initiate CNTRL-C (and TERM) handler  */
  
  if(signal(SIGINT, cbrk_handler) == SIG_ERR) {
    PERROR("Signal Failed (SIGINT).");
    EXIT(0);
  }
  if(signal(SIGTERM, cbrk_handler) == SIG_ERR) {
    PERROR("Signal Failed (SIGTERM).");
    EXIT(0);
  }
  
  /*
    printf("\n  LMT 210 GHz SITE SURVEY RADIOMETER DATA LOGGER REV. 3.0\n");
    printf("  University of Massachusetts, FCRAO @ 1994\n");
  */
  
  firtim = 0;
  
  /*============================          Setup the com port */
  
  /* Open Serial Port :  8 bits data, 1 bit stop, No parity, 2400 baud */
  
  if ((fps = fopen(TIPPER_DEV,"r")) == NULL){
    PERROR("Cannot open serial port for reading.");
    EXIT(1);
  } 
  if ((fpo = fopen(TIPPER_DEV,"w")) == NULL){
    PERROR("Cannot open serial port for writing.");
    EXIT(1);
  }
  
  /*  Set port to 2400 baud */
 
  if( ioctl(fileno(fps), TCGETA, &term) == -1) {
    PERROR("Can not open serial port. Not a TTY");
    EXIT(1);
  }
  
  /*  Save old tty state  */
  
  save = term;
  
  /*  Zero out the baud rate portion of c_cflag  */
  
  term.c_cflag &= ~CBAUD;
  
  /*  Set baud rate to 2400  */
  
  term.c_cflag |= B2400;
  term.c_cflag &= ~(CSIZE|PARENB);
  term.c_cflag |= CS8 ;
  
  ioctl(fileno(fps), TCSETAW, &term);
  ioctl(fileno(fpo), TCSETAW, &term);
  
  /*  Turn of Canonical Processing and echo */
  
  term.c_lflag &= ~ICANON;
  term.c_lflag &= ~ECHO;
  
  /*  Set MIN to one and TIME to zero  */
  
  term.c_cc[VMIN]  = 1;
  term.c_cc[VTIME] = 0;
  
  /*  Disable NL to CR mapping */  
  
  term.c_iflag &= ~ICRNL;
  
  /*  Set new terminal state  */
  
  ioctl(fileno(fps), TCSETA, &term);
  ioctl(fileno(fpo), TCSETA, &term);
  
  /*===========================          End  Setup com port */
  
  daysent =0;
  for(;;){	
    
    /*  CNTRL-C the radiometer and gets its attention */
    PVERB("[CNTRL-C the radiometer]\n");
    PVERB("[Wait for prompt (>)]\n");
    fprintf(fpo, "%c", toascii(3));
    fflush(fpo);
    do{	
      READ(fileno(fps), &c, sizeof(char));   
      /* write(1, &c, sizeof(char));*/
    }while(c != '>');
    
    /*  Send RUN command and start the radiometer firmware  */
    PVERB("[Send RUN command]\n");
    fprintf(fpo,"RUN\r\n");  
    
    /*  Wait for firmware prompt (?)  */
    PVERB("[Wait for firmware prompt (?)]\n");
    
    do{
      READ(fileno(fps), &c, sizeof(char));
      /*	    putchar(c);*/
    }while(c != '?');
    
    /*  Get time and Julian date, send it to the radiometer  */
    
    time(&now);
    tmstruct = localtime(&now);
    hh = tmstruct->tm_hour;
    min = tmstruct->tm_min;
    sec = tmstruct->tm_sec;
    
    mm = tmstruct->tm_mon+1;
    dd = tmstruct->tm_mday;
    yy = tmstruct->tm_year+1900;
    
    /* printf("%d\n",yy);*/
    refday=julday(mm,dd,yy)-julday(1,1,1992)+1L;
    PVERB("[%d,%d,%d,%ld]\n", hh, min, sec, refday);
    fprintf(fpo, "%d,%d,%d,%ld\r\n", hh, min, sec, refday); 
    /* printf( " %d,%d,%d,%d\n", hh, min, sec, refday); */
    
    /*  Wait for the radiometer to home the mirror, wait for prompt (C) */
    PVERB("[Wait for the radiometer to home the mirror]\n");
    PVERB("[wait for prompt (C)]\n");
    /* 	printf("\n  Initializing Mirror Position... \n ");*/
    do {
      READ(fileno(fps), &c,sizeof(char));
      /*          printf("%c",c);*/
    } while (c != 'C');
    
    /* 	printf("\nDone.\n\n");*/
    
    /* Create date stamp and open data file */
    time(&now);
    tmstruct = localtime(&now);
    mm = tmstruct->tm_mon+1;
    dd = tmstruct->tm_mday;
    yy = tmstruct->tm_year;
    yy -= 100;
    
    if(name != NULL) free(name);
    name = (char *) malloc(sizeof(char)*(strlen(data_dir)+9));
    if (!name) { PERROR("malloc()"); EXIT(1); }
    sprintf(name,"%s/su%.2d%.2d%.2d", data_dir, yy, mm, dd);
    PVERB("[Data file: %s]\n",name);
    SYSLOG(LOG_NOTICE,"Open data file %s",name);
    /*	printf("  Data File: %s\n\n",name);*/
    if((fp = fopen(name,"a")) == NULL){
      /*	printf("   ERROR: Can not open file %s\n",name); */
      PERROR("Can not data open file."); 
      fclose(fps);
      EXIT(0);
    }
    /* If this is the first time through loop send */
    /* the radiometer A command to start data taking */
    
    do{
      READ(fileno(fps), &c, sizeof(char));
      /*putchar(c);*/
    }while(c != '?');
    
    if(firtim == 0){
      firtim=1;
      PVERB("[Send 'A' command for start data taking]\n");
      fprintf(fpo,"A\r\n");
    }
    /* Write data to file until the day changes */
    do {	   
      READ(fileno(fps), &c, sizeof(char));
      /* write(1, &c, sizeof(char)); */
      if(c != toascii(13)) write(fileno(fp), &c, sizeof(char));
      /* Write buffer to disk at begining of a scan */
      if (c == '*') fflush(fp);
      time(&now);
      tmstruct = localtime(&now);
      day = tmstruct->tm_mday;
    }while(dd==day);
    fflush(fp);
    SYSLOG(LOG_NOTICE,"Close data file %s",name);
    fclose(fp);
    firtim=0;
  } 
  return 0;
}

/********************************************/
/*****  Function: Calculate Julian Day  *****/
/********************************************/

long julday(int mm, int id, int iyyy)
{
  long jul;
  int ja,jy,jm;
  
  if(iyyy==0) PERROR("0 bad Julian Day");
  if(iyyy<0) ++iyyy;
  if(mm>2){
    jy=iyyy;
    jm=mm+1;
  }
  else {
    jy=iyyy-1;
    jm=mm+13;
  }
  jul=(long)(floor(365.25*jy)+floor(30.6001*jm)+id+1720995);
  if(id+31L*(mm+12L*iyyy) >= IGREG ){
    ja=0.01*jy;
    jul+=2-ja+(int)(0.25*ja);
  }
  return(jul);
}

/***********************************************/
/*****  Function: Break Handler (CNTRL-C)  *****/
/*****               and TERM              *****/
/***********************************************/

void cbrk_handler(int s)
{
  if (fp != NULL) {
    fflush(fp);
    fclose(fp);
    SYSLOG(LOG_NOTICE,"Close data file on signal %d",s);
  }
  /*  CNTRL-C the radiometer */
  fprintf(fpo, "%c", toascii(3));
  fflush(fpo);
  if (fps != NULL) fclose(fps);
  if (fpo != NULL) fclose(fpo);
  ioctl(fileno(fps), TCSETA, &save);
  PVERB("Closing %s on signal %d.\n",program_name,s);
  SYSLOG(LOG_NOTICE,"Exit on signal %d",s);
  EXIT(0);
  /* signal(SIGINT, cbrk_handler); ??? */
}

